/****************************************************************************
 * The Fifth ICFP Programming Contest entry
 * Copyright (C) 2002  Dragan Tubic
 * tdragan@gel.ulaval.ca
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ****************************************************************************/

#include "brain.h"
#include <math.h>
#include <stdlib.h>
#include <time.h>

#define KILL_BID 100

//---------------------------------------------------------------------------
Brain::Brain()
{

}
//---------------------------------------------------------------------------
Brain::Brain( string hostname, int port )
{
    Init( hostname, port );
}
//---------------------------------------------------------------------------
void Brain::Init( string hostname, int port )
{
    time_t t;
    srand( (unsigned) time(&t) );
    mServer.Connect( hostname, port );
    mServer.Init( mMap, mRobots, mSelfId, mPackages );
    mServer.UpdateRobots( mRobots, mPackages, mPickedPackages );
    // Should be zero ALWAYS
    mSelfIndex = mServer.FindRobot( mRobots, mSelfId );

    mTurnCounter = 0;
    mGoal = undefined_goal;

    SelectNewGoal();
}
//---------------------------------------------------------------------------
void Brain::GoBaby()
{
}
//---------------------------------------------------------------------------
void Brain::SelectNewGoal()
{
    mProgress = 0;
    mProgressCounter = 0;
    mLastDistance = -1;

    // Select a longterm goal
    int goal_x = -1;
    int goal_y = -1;
    if ( ( mGoal == undefined_goal ) || ( mPickedPackages.size() == 0 ) )
    {
        // set the closest non-empty home base as the goal
        int closest_base = mMap.FindClosestHomeBase( mRobots[0].x,
                                                     mRobots[0].y );
        if ( closest_base >= 0 )
        {
            goal_x = mMap.GetHomeBase(closest_base).x;
            goal_y = mMap.GetHomeBase(closest_base).y;
            mGoal = pickup_goal;
        } else
        {
            // No more packages left at the bases
            // go search packages at random
            // I seriously think it is not a good idea
            if ( mGoal == search_goal )
//            if ( true )
            {
                // Check the bases again
                for ( unsigned int i = 0; i < mMap.NoHomeBases(); i++ )
                {
                    mMap.GetHomeBase(i).empty = false;
                }
                int closest_base = mMap.FindClosestHomeBase( mRobots[ mSelfIndex ].x,
                                                             mRobots[ mSelfIndex ].y );
                if ( closest_base >= 0 )
                {
                    goal_x = mMap.GetHomeBase(closest_base).x;
                    goal_y = mMap.GetHomeBase(closest_base).y;
                    mGoal = pickup_goal;
                }
            } else
            {
                mGoal = search_goal;
                SquareType type = wall;
                while ( type != empty )
                {
                    goal_x = ( rand() % ( mMap.SizeX() - 1 ) ) + 1;
                    goal_y = ( rand() % ( mMap.SizeY() - 1 ) ) + 1;
                    type = (SquareType)mMap(goal_x,goal_y).mType;
                }
            }
        }
    }
    if ( mPickedPackages.size() > 0 )
    {
        // Go deliver!
        int no_pack = mPickedPackages.size();

/*
        // Find the package closest to the average position
        // of all picked packages
        int mean_x = 0;
        int mean_y = 0;
        for ( unsigned int i = 0; i < no_pack; i++ )
        {
            mean_x += mPickedPackages[i].dest_x;
            mean_y += mPickedPackages[i].dest_y;
        }
        mean_x /= no_pack;
        mean_y /= no_pack;
        int sel_x = mean_x;
        int sel_y = mean_y;
*/
        // or
        int sel_x = mRobots[0].x;
        int sel_y = mRobots[0].y;

        int closest = -1;
        int min_dist = 10000;
        for ( unsigned int i = 0; i < no_pack; i++ )
        {
            int dist = abs( sel_x - mPickedPackages[i].dest_x ) +
                       abs( sel_y - mPickedPackages[i].dest_y );
            if ( dist < min_dist )
            {
                closest = i;
                min_dist = dist;
            }
        }

        goal_x = mPickedPackages[closest].dest_x;
        goal_y = mPickedPackages[closest].dest_y;
        mGoal = delivery_goal;
    }
    mGoalX = goal_x;
    mGoalY = goal_y;
    mMap.ComputeGoalDistances( goal_x, goal_y );
}
//---------------------------------------------------------------------------
void Brain::SelectRandomSearchGoal()
{
    mProgress = 0;
    mProgressCounter = 0;
    mLastDistance = -1;

    mGoal = search_goal;
    SquareType type = wall;
    while ( type != empty )
    {
        mGoalX = ( rand() % ( mMap.SizeX() - 1 ) ) + 1;
        mGoalY = ( rand() % ( mMap.SizeY() - 1 ) ) + 1;
        type = (SquareType)mMap(mGoalX,mGoalY).mType;
    }
    mMap.ComputeGoalDistances( mGoalX, mGoalY );
}
//---------------------------------------------------------------------------
int Brain::FindRobot( int x, int y )
{
    for ( unsigned int i = 1; i < mRobots.size(); i++ )
    {
        if ( ( mRobots[i].x == x ) && ( mRobots[i].y == y ) )
        {
            return i;
        }
    }
    return -1;
}
//---------------------------------------------------------------------------
int Brain::FindRobotPred( int x, int y )
{
    for ( unsigned int i = 1; i < mRobots.size(); i++ )
    {
        if ( ( mRobots[i].x + mRobots[i].pred_x == x ) &&
             ( mRobots[i].y + mRobots[i].pred_y == y ) )
        {
            return i;
        }
    }
    return -1;
}
//---------------------------------------------------------------------------
int Brain::FindPackageByDest( Packages rPackages, int x, int y )
{
    for ( unsigned int i = 0; i < rPackages.size(); i++ )
    {
        if ( ( rPackages[i].dest_x == x ) && ( rPackages[i].dest_y == y ) )
        {
            return i;
        }
    }
    return -1;
}
//---------------------------------------------------------------------------
void Brain::EvaluateStep( Step& step, bool loaded )
{
    int x = step.dx + mRobots[0].x;
    int y = step.dy + mRobots[0].y;
    if ( ( x < 1 ) ||
         ( x > mMap.SizeX() ) ||
         ( y < 1 ) ||
         ( y > mMap.SizeY() ) )
    {
        step.lethal = false;
        step.movable = false;
        step.progress = -10;
        step.overall = -10;
        step.can_kill = false;
        return;
    }

    // is there a lethal field
    step.lethal = ( mMap(x,y).mType == lethal );
    // is there a wall
    step.movable = ( mMap(x,y).mType == empty );
    // are we approaching the goal and how much
    int curr_goal_dist = mMap(mRobots[0].x,mRobots[0].y).mGoalDistance;
    int curr_wall_dist = mMap(mRobots[0].x,mRobots[0].y).mWallDistance;
    step.progress = mMap.EvaluateStep( x, y, curr_goal_dist, curr_wall_dist );

    // is there a robot
    step.can_kill = false;
    int rob_ind = FindRobot(x,y);
    step.robot = (rob_ind > 0);
    int rob_ind_pred = FindRobotPred(x,y);
    step.robot_pred = (rob_ind_pred > 0);
    if ( step.robot )
    {
        step.can_kill = false;
        // can we kill it?
        int rx = x + step.dx;
        int ry = y + step.dy;
        if ( ( rx < 1 ) ||
             ( rx > mMap.SizeX() ) ||
             ( ry < 1 ) ||
             ( ry > mMap.SizeY() ) )
        {
            step.can_kill = (mMap(rx,ry).mType == lethal);
        }
    }
    // Can we deliver something here?
    int pack_ind = FindPackageByDest(mPickedPackages,x,y);
    step.deliverable = (pack_ind >= 0);

    step.overall = step.progress;
    if ( step.deliverable ) step.overall += 2;
    if ( loaded && step.robot )
    {
        step.overall -= 2;
        if ( step.overall < 0 ) step.overall = 0;
    }
    if ( loaded && step.robot_pred )
    {
        step.overall -= 2;
        if ( step.overall < 0 ) step.overall = 0;
    }
}
//---------------------------------------------------------------------------
int Brain::ClusterPackages()
{
    // Make 16 groups ie divide the map into 16 squares
    int size_x = mMap.SizeX() / 4;
    int size_y = mMap.SizeY() / 4;
    int sizes[16];
    for ( int i = 0; i < 16; i++ ) sizes[i] = 0;
    int max = -1;
    int max_index = -1;
    for ( unsigned int i = 0; i < mPackages.size(); i++ )
    {
        int x = mPackages[i].dest_x / size_x;
        int y = mPackages[i].dest_y / size_y;
        int index = 4*y + x;
        sizes[index]++;
        if ( sizes[index] > max )
        {
            max = sizes[index];
            max_index = index;
        }
    }

    int min_dist = 10000;
    int robo_x = mRobots[0].x / size_x;
    int robo_y = mRobots[0].y / size_y;
    // Among the groups of max_sizes find the closest one
    int chk = max_index;
    for ( unsigned int i = 0; i < 16; i++ )
    {
        if ( ( abs( sizes[i] - max ) <= 1 ) && sizes[i] > 1 )
        {
            int x = i % 4;
            int y = i / 4;
            int dist = abs(x-robo_x) + abs(y-robo_y);
            if ( min_dist > dist )
            {
                max_index = i;
                min_dist = dist;
            }
        }
    }

    if ( max_index > 0 )
    {
        int dbg = 0;
    }
    return max_index;
}
//---------------------------------------------------------------------------
bool Brain::PickupPackage( int bid )
{
    int package_group = ClusterPackages();

    if ( package_group == -1 ) return false;
    // How much do we carry?
    int load = 0;
    for ( unsigned int i = 0; i < mPickedPackages.size(); i++ )
    {
        load += mPickedPackages[i].weight;
    }

    int capacity = mRobots[0].capacity;
    // What can we load
    vector<int> pack_id;
    int new_load = load;
    bool found = false;

    int group_size_x = mMap.SizeX() / 4;
    int group_size_y = mMap.SizeY() / 4;

    for ( unsigned int i = 0; i < mPackages.size(); i++ )
    {
        int x = mPackages[i].dest_x / group_size_x;
        int y = mPackages[i].dest_y / group_size_y;
        int index = 4*y + x;
        if ( index != package_group ) continue;

        int weight = mPackages[i].weight;
        if ( capacity >= new_load + weight )
        {
            found = true;
            pack_id.push_back( mPackages[i].id );
            new_load += weight;
        }
    }
    if ( !found )
    {
        // Nothing to pickup, move on
        return false;
    }

    strstream command_stream;
    command_stream << bid << " Pick ";
    for ( unsigned int i = 0; i < pack_id.size(); i++ )
    {
        command_stream << pack_id[i] << " ";
    }
    command_stream << "\n" << ends;
    string command = command_stream.str();
    mServer.ExecuteCommand( command );
    return true;
}
//---------------------------------------------------------------------------
bool Brain::DeliverPackage( int bid )
{
    int x = mRobots[0].x;
    int y = mRobots[0].y;
    vector<int> pack_id;
    bool found = false;
    for ( unsigned int i = 0; i < mPickedPackages.size(); i++ )
    {
        if ( ( mPickedPackages[i].dest_x == x ) &&
             ( mPickedPackages[i].dest_y == y ) )
        {
            pack_id.push_back( mPickedPackages[i].id );
            found = true;
        }
    }
    if ( !found )
    {
        // Nothing to deliver move on
        return false;
    }

    strstream command_stream;
    command_stream << bid << " Drop ";
    for ( unsigned int i = 0; i < pack_id.size(); i++ )
    {
        command_stream << pack_id[i] << " ";
    }
    command_stream << "\n" << ends;
    string command = command_stream.str();
    mServer.ExecuteCommand( command );
    return true;
}
//---------------------------------------------------------------------------
void Brain::Move( string& dir, int bid )
{
    strstream command_stream;
    command_stream.clear();
    command_stream << bid << " Move " << dir << "\n" << ends;
    string command = command_stream.str();
    mServer.ExecuteCommand( command );
}
//---------------------------------------------------------------------------
void Brain::MakeStep()
{
    string WS = "W";
    string ES = "E";
    string SS = "S";
    string NS = "N";
    // Priorities:
    //  1. Avoid danger
    //  2. Kill oponent
    //  3. Deliver
    //  4. Pickup
    //  5. Move
    //
    // Do we approach the goal, or we are stacked with some
    // braindead robot?
    int x = mRobots[0].x;
    int y = mRobots[0].y;
    int dist = mMap(x,y).mGoalDistance;
    if ( mLastDistance > -1 )
    {
        mProgress += ( mLastDistance - dist );
        mLastDistance = dist;
    } else
    {
        mLastDistance = dist;
    }
    mDistanceCheck[mProgressCounter] = dist;
    mProgressCounter++;
    if ( mProgressCounter > 10 )
    {
        float stack_measure = (float)mProgress;
        stack_measure /= (float)mProgressCounter;
        mProgressCounter = 0;
        mProgress = 0;
        mProgressCounter = 0;
        mLastDistance = -1;

        if ( stack_measure < 0.2 )
        {
            // Yes we are stacked
            // select a random search goal
            SelectRandomSearchGoal();
        } else
        {
            if ( mGoal == search_goal ) SelectNewGoal();
        }
    }

    // Are we at the goal?
    bool at_goal = false;
    if ( ( mGoalX == mRobots[0].x ) && ( mGoalY == mRobots[0].y ) )
    {
        at_goal = true;
    }

    // Are we loaded?
    bool loaded = ( mPickedPackages.size() );

    Step west, east, north, south;
    west.dx  = -1; west.dy  =  0; west.dir = WS;
    east.dx  =  1; east.dy  =  0; east.dir = ES;
    north.dx =  0; north.dy =  1; north.dir = NS;
    south.dx =  0; south.dy = -1; south.dir = SS;
    EvaluateStep( west, loaded );
    EvaluateStep( east, loaded );
    EvaluateStep( north, loaded );
    EvaluateStep( south, loaded );

    // Are we in a dangerour position?
    //    Is there a lethal square nearby and a robot on the oposite side of the lethal field?
    // kill or avoid if possible
    if ( east.robot && west.lethal )
    {
        // Can we kill it?
        if ( east.can_kill ) Move(ES,KILL_BID);

        // No escape, push the bastard
        if ( south.lethal && north.lethal ) Move(ES,KILL_BID);

        // Evade
        if ( south.overall > north.overall )
        {
            if ( south.movable && !south.lethal )
            {
                Move(SS,KILL_BID);
                return;
            }
        } else
        {
            if ( north.movable && !north.lethal )
            {
                Move(NS,KILL_BID);
                return;
            }
        }
        // No escape, push the bastard
        Move(ES,KILL_BID);
        return;
    }
    if ( west.robot && east.lethal )
    {
        // Can we kill it?
        if ( west.can_kill ) Move(WS,KILL_BID);

        // No escape, push the bastard
        if ( south.lethal && north.lethal ) Move(WS,KILL_BID);

        // Evade
        if ( south.overall > north.overall )
        {
            if ( south.movable && !south.lethal )
            {
                Move(SS,KILL_BID);
                return;
            }
        } else
        {
            if ( north.movable && !north.lethal )
            {
                Move(NS,KILL_BID);
                return;
            }
        }
        // No escape, push the bastard
        Move(WS,KILL_BID);
        return;
    }
    if ( south.robot && north.lethal )
    {
        // Can we kill it?
        if ( south.can_kill ) Move(SS,KILL_BID);

        // No escape, push the bastard
        if ( east.lethal && west.lethal ) Move(SS,KILL_BID);

        // Evade
        if ( west.overall > east.overall )
        {
            if ( west.movable && !west.lethal )
            {
                Move(WS,KILL_BID);
                return;
            }
        } else
        {
            if ( east.movable && !east.lethal )
            {
                Move(ES,KILL_BID);
                return;
            }
        }

        // No escape, push the bastard
        Move(SS,KILL_BID);
        return;
    }
    if ( north.robot && south.lethal )
    {
        // Can we kill it?
        if ( north.can_kill ) Move(NS,KILL_BID);

        // No escape, push the bastard
        if ( east.lethal && west.lethal ) Move(NS,KILL_BID);

        // Evade
        if ( west.overall > east.overall )
        {
            if ( west.movable && !west.lethal )
            {
                Move(WS,KILL_BID);
                return;
            }
        } else
        {
            if ( east.movable && !east.lethal )
            {
                Move(ES,KILL_BID);
                return;
            }
        }

        // No escape, push the bastard
        Move(NS,KILL_BID);
        return;
    }

    // Kill oponent
    if ( east.can_kill )
    {
        Move(ES,KILL_BID);
        return;
    }
    if ( west.can_kill )
    {
        Move(WS,KILL_BID);
        return;
    }
    if ( south.can_kill )
    {
        Move(SS,KILL_BID);
        return;
    }
    if ( north.can_kill )
    {
        Move(NS,KILL_BID);
        return;
    }

    // Are there robots around?
    bool robots = east.robot || west.robot || south.robot || north.robot;
    int bid = 1;
    if ( robots ) bid = 12;

    // Can we deliver something here?
    if ( mPickedPackages.size() > 0 )
    {

        if ( DeliverPackage( bid ) ) return;
        // Nothing delivered, continue
    }

    // Can we pickup something here?
    if ( mGoal != search_goal )
    {
        if ( PickupPackage( bid ) )
        {
            return;
            // if full do something else
        }
    }

    // Find the best move
    float best_overall = east.overall;
    string best_dir = ES;

    if ( west.overall > best_overall )
    {
        best_dir = WS;
        best_overall = west.overall;
    }
    if ( south.overall > best_overall )
    {
        best_dir = SS;
        best_overall = south.overall;
    }
    if ( north.overall > best_overall )
    {
        best_dir = NS;
        best_overall = north.overall;
    }

    if ( east.overall == best_overall )
    {
        if ( rand() & 1 == 1 ) best_dir = ES;
    }
    if ( west.overall == best_overall )
    {
        if ( rand() & 1 == 1 ) best_dir = WS;
    }
    if ( south.overall == best_overall )
    {
        if ( rand() & 1 == 1 ) best_dir = SS;
    }
    if ( north.overall == best_overall )
    {
        if ( rand() & 1 == 1 ) best_dir = NS;
    }



    if ( at_goal && ( mGoal == pickup_goal ) && ( mPackages.size() == 0 ) )
    {
        // base empty
        int ind = mMap.FindClosestHomeBase(x,y);
        if ( ind >= 0 )
        {
            mMap.GetHomeBase(ind).empty = true;
        }
    }

    if ( at_goal )
    {
        SelectNewGoal();
    }

    if ( !loaded && mGoal == delivery_goal )
    {
        // We have lost packages change plan
        SelectNewGoal();
    }


    Move(best_dir,1);
}
//---------------------------------------------------------------------------

