/****************************************************************************
 * The Fifth ICFP Programming Contest entry
 * Copyright (C) 2002  Dragan Tubic
 * tdragan@gel.ulaval.ca
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ****************************************************************************/

#ifndef __BRAIN_H__
#define __BRAIN_H__
#include "contest.h"
#include <string>
//---------------------------------------------------------------------------
using namespace std;
//---------------------------------------------------------------------------
enum GoalType { undefined_goal = -1, pickup_goal = 0, delivery_goal = 1, search_goal = 2 };
//---------------------------------------------------------------------------
class Step
{
public:
    // is there a lethal field
    bool  lethal;
    // is there a wall
    bool  movable;
    // is there a robot
    bool  robot;
    // is there a chance that robot might come here
    bool  robot_pred;
    // direction
    int   dx;
    int   dy;
    // is there a killable robot
    bool  can_kill;
    // are we approaching the goal and how much
    float progress;
    // Can we deliver something here?
    bool deliverable;
    // Overall
    float overall;
    // direction
    string dir;
};
//---------------------------------------------------------------------------
class Brain
{
public:
    Brain();
    Brain( string hostname, int port );
    void Init( string hostname, int port );
    void GoBaby();
    void SelectNewGoal();
    void MakeStep();
    void EvaluateStep( Step& step, bool loaded );
    int FindRobot( int x, int y );
    int FindRobotPred( int x, int y );    
    int FindPackageByDest( Packages rPackages, int x, int y );
    bool PickupPackage( int bid );
    bool DeliverPackage( int bid );
    void Move( string& dir, int bid );
    void SelectRandomSearchGoal();
    int ClusterPackages();
//private:
    WorldMap mMap;
    Server   mServer;
    Robots   mRobots;
    Packages mPackages;
    Packages mPickedPackages; 
    int      mSelfIndex;
    int      mSelfId;
    int      mTurnCounter;
    GoalType mGoal;
    int      mGoalX;
    int      mGoalY;

    int     mProgress;
    int     mProgressCounter;
    int     mLastDistance;
    int     mDistanceCheck[15];
};
//---------------------------------------------------------------------------
#endif // __BRAIN_H__
