/****************************************************************************
 * The Fifth ICFP Programming Contest entry
 * Copyright (C) 2002  Dragan Tubic
 * tdragan@gel.ulaval.ca
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ****************************************************************************/

#include "contest.h"
//---------------------------------------------------------------------------
#include <stdexcept>
#include <string>
#include <vector>
#include <sstream>
#include <iostream>
//---------------------------------------------------------------------------
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <fcntl.h>
//---------------------------------------------------------------------------
#include <unistd.h>
//---------------------------------------------------------------------------
using namespace std;
//---------------------------------------------------------------------------
HomeBase::HomeBase()
{
    x = -1;
    y = -1;
    empty = false;
}
//---------------------------------------------------------------------------
Square::Square()
{
    mGoalDistance = -1;
    mWallDistance = -1;
    mType = -1;
}
//---------------------------------------------------------------------------
Robot::Robot()
{
    pred_x = 0;
    pred_y = 0;
    x = -1;
    y = -1;
    id = -1;
    capacity = -1;
    money = -1;
}
//---------------------------------------------------------------------------
Square& WorldMap::operator ()(int x, int y)
{
    if ( ( x < 1 ) || ( x > mMapSizeX ) ||
         ( y < 1 ) || ( y > mMapSizeY ) )
    {
        throw std::runtime_error( string("Index of WorldMap(i,j) out of range.") );
        return mMap[0][0];
    }
    return mMap[y-1][x-1];
}
//---------------------------------------------------------------------------
HomeBase& WorldMap::GetHomeBase( int index )
{
    return mHomeBases[index];
}
//---------------------------------------------------------------------------
int WorldMap::NoHomeBases()
{
    return mHomeBases.size();
}
//---------------------------------------------------------------------------
int WorldMap::FindClosestHomeBase( int x, int y )
{
    int min_dist = 10000;
    int closest = -1;
    for ( unsigned int i = 0; i < mHomeBases.size(); i++ )
    {
        int dist = abs( mHomeBases[i].x - x ) +
                   abs( mHomeBases[i].y - y );
        if ( (dist < min_dist) && !mHomeBases[i].empty )
        {
            closest = i;
            min_dist = dist;
        }
    }
    return closest;
}
//---------------------------------------------------------------------------
WorldMap::WorldMap()
{
    mMap = NULL;
}
//---------------------------------------------------------------------------
WorldMap::WorldMap( unsigned int sizeX, unsigned int sizeY )
{
    mMapSizeX = sizeX;
    mMapSizeY = sizeY;
    AllocateMemory();
}
//---------------------------------------------------------------------------
void WorldMap::NewMap( unsigned int sizeX, unsigned int sizeY )
{
    mHomeBases.clear();
    ReleaseMemory();
    mMapSizeX = sizeX;
    mMapSizeY = sizeY;
    AllocateMemory();
}
//---------------------------------------------------------------------------
void WorldMap::NewMap( Server& rServer )
{
    int size_x, size_y;
    string tmp_string;
    strstream tmp_stream;

    rServer.ReadLine( tmp_string );
    tmp_stream << tmp_string.c_str();
    tmp_stream >> size_x >> size_y;

    NewMap( size_x, size_y );
    char c;
    for ( unsigned int y = 1; y <= size_y; y++ )
    {
        rServer.ReadLine( tmp_string );
        tmp_stream.clear();
        tmp_stream << tmp_string.c_str();

        for ( unsigned int x = 1; x <= size_x; x++ )
        {
            HomeBase base;
            SquareType type = invalid;
            c = 0;
            while ( c < 33 ) tmp_stream >> c;
            switch (c)
            {
              case '.': type = empty;
              break;
              case '#': type = wall;
              break;
              case '~': type = lethal;
              break;
              case '@': type = empty;
                        base.x = x;
                        base.y = y;
                        base.empty = false;
                        mHomeBases.push_back(base);
              break;
              default: throw runtime_error( "Unknow square type" );
            }
            (*this)(x,y).mType = type;

        }
    }
}
//---------------------------------------------------------------------------
WorldMap::~WorldMap()
{
    ReleaseMemory();
}
//---------------------------------------------------------------------------
void WorldMap::AllocateMemory()
{
    mMap = new Square*[mMapSizeY];
    for ( unsigned int i = 0; i < mMapSizeY; i++ )
    {
        mMap[i] = new Square[mMapSizeX];
    }
}
//---------------------------------------------------------------------------
void WorldMap::ReleaseMemory()
{
    if ( mMap == NULL ) return;
    for ( unsigned int i = 0; i < mMapSizeY; i++ )
    {
        delete [] mMap[i];
    }
    delete [] mMap;
}
//---------------------------------------------------------------------------
int WorldMap::FindClosestWall( int posX, int posY )
{
    int max_dist = min(posX-1,posY-1);
    max_dist = min(max_dist,mMapSizeX-posX);
    max_dist = min(max_dist,mMapSizeY-posY);

    if ( max_dist == 0 ) return 1;

    for ( int dst = 1; dst <= max_dist; dst++ )
    {
        int min_x = posX - dst;
        int max_x = posX + dst;

        int min_y = posY - dst;
        int max_y = posY + dst;

        for ( int d = 1; d <= dst; d++ )
        {
            if ( (*this)(min_x+d-1,posY+d-1).mType == wall ) return dst;
            if ( (*this)(min_x+d,posY-d).mType == wall ) return dst;
            if ( (*this)(max_x-d,posY+d).mType == wall ) return dst;
            if ( (*this)(max_x-d+1,posY-d+1).mType == wall ) return dst;
        }
    }
    return max_dist+1;
}
//---------------------------------------------------------------------------
void WorldMap::ComputeWallDistances()
{
    for ( unsigned int x = 1; x <= mMapSizeX; x++ )
    {
        for ( unsigned int y = 1; y <= mMapSizeY; y++ )
        {
            if ( (*this)(x,y).mType == empty )
            {
                (*this)(x,y).mWallDistance = FindClosestWall(x,y);
            }
        }
    }
}
//---------------------------------------------------------------------------
#define SET_GOAL_DIST(x,y) \
    if ( ( (*this)(x,y).mGoalDistance == -1 ) &&\
         ( (*this)(x,y).mType == empty ) ) \
    { \
        (*this)(x,y).mGoalDistance = currDist; \
        new_points_x[point_counter] = x; \
        new_points_y[point_counter] = y; \
        point_counter++; \
    } \

//---------------------------------------------------------------------------
void WorldMap::DoNeighbours( int currDist,
                             int ptX,
                             int ptY,
                             int* new_points_x,
                             int* new_points_y,
                             int& point_counter )
{
    int left_x = ptX - 1;
    if ( ptX == 1 ) left_x = 1;
    int right_x = ptX + 1;
    if ( ptX == mMapSizeX ) right_x = mMapSizeX;
    int bot_y = ptY - 1;
    if ( ptY == 1 ) bot_y = 1;
    int up_y = ptY + 1;
    if ( ptY == mMapSizeY ) up_y = mMapSizeY;

//    SET_GOAL_DIST(left_x,up_y)
    SET_GOAL_DIST(left_x,ptY)
//    SET_GOAL_DIST(left_x,bot_y)

//    SET_GOAL_DIST(right_x,up_y)
    SET_GOAL_DIST(right_x,ptY)
//    SET_GOAL_DIST(right_x,bot_y)

    SET_GOAL_DIST(ptX,up_y)
    SET_GOAL_DIST(ptX,bot_y)
}
//---------------------------------------------------------------------------
void WorldMap::ComputeGoalDistances( int goalX, int goalY )
{
    // reset all distances
    for ( unsigned int x = 1; x <= mMapSizeX; x++ )
    {
        for ( unsigned int y = 1; y <= mMapSizeY; y++ )
        {
            (*this)(x,y).mGoalDistance = -1;
        }
    }

    // OK we start with a single point, a goal
    int* current_pts_x = new int[1024];
    int* current_pts_y = new int[1024];
    int* new_pts_x = new int[1024];
    int* new_pts_y = new int[1024];

    int new_point_counter = 0;
    int current_point_counter = 1;
    current_pts_x[0] = goalX;
    current_pts_y[0] = goalY;
    (*this)(goalX,goalY).mGoalDistance = 0;

    int current_distance = 1;

    while ( current_point_counter > 0 )
    {
        for ( unsigned int i = 0; i < current_point_counter; i++ )
        {
            DoNeighbours( current_distance,
                          current_pts_x[i],
                          current_pts_y[i],
                          new_pts_x,
                          new_pts_y,
                          new_point_counter );
        }
        current_point_counter = new_point_counter;

        int* tmp_ptr = current_pts_x;
        current_pts_x = new_pts_x;
        new_pts_x = tmp_ptr;

        tmp_ptr = current_pts_y;
        current_pts_y = new_pts_y;
        new_pts_y = tmp_ptr;

        new_point_counter = 0;

        current_distance++;
    }

    delete [] current_pts_x;
    delete [] current_pts_y;
    delete [] new_pts_x;
    delete [] new_pts_y;
}
//---------------------------------------------------------------------------
float WorldMap::EvaluateStep( int x, int y, int currGoalDist, int currWallDist )
{
    int d_goal_dist = (*this)(x,y).mGoalDistance - currGoalDist;
    int d_wall_dist = (*this)(x,y).mWallDistance;
    int type = (*this)(x,y).mType;
    if ( type == empty )
    {
        float tmp_quality = -d_goal_dist + 1;
        tmp_quality += d_wall_dist/2.0f;
        if ( d_wall_dist == 1 )
        {
            tmp_quality += 0;
        }
        return tmp_quality;
    }
    return -1.0f;
}
//---------------------------------------------------------------------------
void WorldMap::MakeStep( int currX, int currY, int& rNewX, int& rNewY )
{
    int current_distance = (*this)(currX,currY).mGoalDistance;
    int current_wall_distance = (*this)(currX,currY).mWallDistance;

    int new_x = currX;
    int new_y = currY;
    float quality = -1;
    // Check north
    if ( currY < mMapSizeY )
    {
        float tmp_quality = EvaluateStep(currX,currY+1,current_distance,current_wall_distance);
        if ( tmp_quality > quality )
        {
            new_x = currX;
            new_y = currY+1;
            quality = tmp_quality;
        }
    }
    // Check south
    if ( currY > 1 )
    {
        float tmp_quality = EvaluateStep(currX,currY-1,current_distance,current_wall_distance);
        if ( tmp_quality > quality )
        {
            new_x = currX;
            new_y = currY-1;
            quality = tmp_quality;
        }
    }
    // Check east
    if ( currX < mMapSizeX )
    {
        float tmp_quality = EvaluateStep(currX+1,currY,current_distance,current_wall_distance);
        if ( tmp_quality > quality )
        {
            new_x = currX+1;
            new_y = currY;
            quality = tmp_quality;
        }
    }
    // Check west
    if ( currX > 1 )
    {
        float tmp_quality = EvaluateStep(currX-1,currY,current_distance,current_wall_distance);
        if ( tmp_quality > quality )
        {
            new_x = currX-1;
            new_y = currY;
            quality = tmp_quality;
        }
    }
    rNewX = new_x;
    rNewY = new_y;
}
//---------------------------------------------------------------------------
void Server::Connect()
{
    mSocket = -1;

	struct sockaddr_in sin;
	struct sockaddr_in pin;
	struct hostent *hp;

	/* go find out about the desired host machine */
	if ( ( hp = gethostbyname( mHostName.c_str() ) ) == 0 )
    {
		throw runtime_error("Gethostbyname error!");
	}

	/* fill in the socket structure with host information */
	memset(&pin, 0, sizeof(pin));
	pin.sin_family = AF_INET;
	pin.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
	pin.sin_port = htons( mHostPort );

	/* grab an Internet domain socket */

	if ( ( mSocket = socket(AF_INET, SOCK_STREAM, 0 ) ) == -1 )
    {
		throw runtime_error("socket error");
	}

	/* connect to PORT on HOST */
	if ( connect( mSocket,(struct sockaddr *)  &pin, sizeof(pin)) == -1)
    {
		throw runtime_error("Connect error");
	}

    int opts;
    opts = fcntl ( mSocket, F_GETFL );
    if ( opts < 0 )
    {
        throw runtime_error("Can't read socket flags.");
    }
    opts = ( opts | O_NONBLOCK );
    fcntl ( mSocket, F_SETFL, opts );
}
//---------------------------------------------------------------------------
void Server::Close()
{
    close( mSocket );
}
//---------------------------------------------------------------------------
Server::Server()
{
}
//---------------------------------------------------------------------------
Server::~Server()
{
    Close();
}
//---------------------------------------------------------------------------
Server::Server( string hostname, int port )
{
    mHostName = hostname;
    mHostPort = port;
    Connect();
}
//---------------------------------------------------------------------------
void Server::Connect( string hostname, int port )
{
    mHostName = hostname;
    mHostPort = port;
    Connect();
}
//---------------------------------------------------------------------------
void Server::Send( string& rMessage )
{
	if ( send( mSocket, rMessage.c_str(), rMessage.size(), 0) == -1)
    {
		throw runtime_error("Error sending a message.");
	}
}
//---------------------------------------------------------------------------
void Server::Receive2( string& rMessage )
{
    /* wait for a message to come back from the server */
    int answer_len = recv( mSocket, mBuffer, BUFFER_SIZE, 0);
    if ( answer_len == -1 )
    {
        throw runtime_error("Error receiving data.");
    }
    mBuffer[answer_len] = 0;
    rMessage = mBuffer;
}
//---------------------------------------------------------------------------
void Server::ReadLine( string& rMessage )
{
    bool line_ready = false;

    int line_len = -1;
    while ( !line_ready )
    {
        int answer_len = recv( mSocket, mBuffer, BUFFER_SIZE, MSG_PEEK );

        if ( answer_len == -1 )
        {
            usleep(10);
        } else
        {
            mBuffer[answer_len] = 0;
            string tmp = mBuffer;
            int new_line_ind = tmp.find( "\n", 0 );
            if ( new_line_ind >= 0 )
            {
                line_ready = true;
                line_len = new_line_ind;
            }
        }
    }
    int answer_len = recv( mSocket, mBuffer, line_len+1, 0 );
    mBuffer[line_len] = 0;
    rMessage = mBuffer;
}
//---------------------------------------------------------------------------
void Server::Init( WorldMap& rMap, Robots& rRobots, int& me, Packages& rPackages )
{
    string server_answer;
    string tmp = "Player\n";
    Send( tmp );


    rMap.NewMap( *this );
    rMap.ComputeWallDistances();

    strstream tmp_stream;
    string tmp_string;

    rRobots.clear();
    ReadLine( tmp_string );
    tmp_stream << tmp_string.c_str();
    // Make my robot also known as me
    Robot my_robo;
    tmp_stream >> my_robo.id >> my_robo.capacity >> my_robo.money;
    rRobots.push_back( my_robo );
    me = my_robo.id;
}
//---------------------------------------------------------------------------
void Server::ExecuteCommand( string& rCommand )
{
    Send( rCommand );
    cout << "Command: " << rCommand << endl;
}
//---------------------------------------------------------------------------
int Server::FindRobot( Robots& rRobots, int id )
{
    for ( unsigned int i = 0; i < rRobots.size(); i++ )
    {
        if ( rRobots[i].id == id ) return i;
    }
    return -1;
}
//---------------------------------------------------------------------------
void Server::CheckRobot( Robots& rRobots, int id )
{
    // Check if robot exists. Create if it doesn't
    if ( FindRobot( rRobots, id ) >= 0 ) return;
    // Not found/ Make a new one
    Robot r;
    r.id = id;
    rRobots.push_back(r);
}
//---------------------------------------------------------------------------
void Server::PositionRobot( Robots& rRobots, int id, int x, int y )
{
    // Check if robot exists. Yell if it doesn't
    int robo_index = FindRobot( rRobots, id );
    if ( robo_index == -1 )
    {
        throw runtime_error("Moving a dead robot!");
    }
    rRobots[robo_index].x = x;
    rRobots[robo_index].y = y;
}
//---------------------------------------------------------------------------
void Server::MoveRobot( Robots& rRobots, int id, int dx, int dy )
{
    // Check if robot exists. Yell if it doesn't
    int robo_index = FindRobot( rRobots, id );
    if ( robo_index == -1 )
    {
        throw runtime_error("Moving a dead robot!");
    }
    rRobots[robo_index].x += dx;
    rRobots[robo_index].y += dy;
    rRobots[robo_index].pred_x = dx;
    rRobots[robo_index].pred_y = dy;
}
//---------------------------------------------------------------------------
bool Server::UpdateRobots( Robots& rRobots, Packages& rPackages, Packages& rPickedPackages )
{
    string tmp_string;
    ReadLine( tmp_string );

    if ( tmp_string[0] != '#' )
    {
        // We are so dead
        return false;
    }
    cout << "Answer: " << tmp_string << endl;
    // OK here we go
    strstream server_stream;
    server_stream << tmp_string.c_str();
    char c;
    Robot* robo;

    vector<int> living_bots;

    int my_id = rRobots[0].id;
    while ( server_stream >> c )
    {
        int pack_id;
        int id;
        int x;
        int y;
        switch (c)
        {
            // Robot info starts
            case '#': server_stream >> id;
                      living_bots.push_back(id);  
                      CheckRobot( rRobots, id );
                break;
            case 'X': server_stream >> x;
                break;
            case 'Y': server_stream >> y;
                      PositionRobot( rRobots, id, x, y );
                break;
            case 'E': ;
                      MoveRobot( rRobots, id, 1, 0 );
                break;
            case 'W': ;
                      MoveRobot( rRobots, id, -1, 0 );
                break;
            case 'S': ;
                      MoveRobot( rRobots, id, 0, -1 );
                break;
            case 'N': ;
                      MoveRobot( rRobots, id, 0, 1 );
                break;
            case 'P': if ( id == my_id )
                      {
                          server_stream >> pack_id;
                          LoadPackage( rRobots, rPackages, rPickedPackages, pack_id );
                      };
                break;
            case 'D': if ( id == my_id )
                      {
                          server_stream >> pack_id;
                          DropPackage( rRobots, rPackages, rPickedPackages, pack_id );
                      };
                break;
          default: ;
        }
    }
    RemoveRobots( rRobots, living_bots );
    return true;
}
//---------------------------------------------------------------------------
void Server::LoadPackage( Robots& rRobots, Packages& rPackages, Packages& rPickedPackages, int packId )
{
    for ( unsigned int i = 0; i < rPackages.size(); i++ )
    {
        if ( rPackages[i].id == packId )
        {
            rPickedPackages.push_back( rPackages[i] );
            return;
        }
    }
}
//---------------------------------------------------------------------------
void Server::DropPackage( Robots& rRobots, Packages& rPackages, Packages& rPickedPackages, int packId )
{
    vector<Package>::iterator i;
    for ( i = rPickedPackages.begin(); i < rPickedPackages.end(); i++ )
    {
        if ( i->id == packId )
        {
            rPickedPackages.erase(i);
            return;
        }
    }
}
//---------------------------------------------------------------------------
void Server::RemoveRobots( Robots& rRobots, vector<int>& rLivingBots )
{
    vector<Robot> new_list;
    for ( unsigned int i = 0; i < rRobots.size(); i++ )
    {
        bool found = false;
        for ( unsigned int j = 0; j < rLivingBots.size(); j++ )
        {
            if ( rLivingBots[j] == rRobots[i].id )
            {
                found = true;
                break;
            }
        }
        if ( found )
        {
            new_list.push_back( rRobots[i] );
        }
        if ( (i==0) && (!found) )
        {
            // Myself index is always 0
            throw runtime_error("I am dead!");
        }
    }
    rRobots = new_list;
}
//---------------------------------------------------------------------------
int Server::UpdatePackages( Packages& rPackages, int currX, int currY )
{
    string server_line;
    ReadLine( server_line );
    cout << "Update packages: " << server_line << endl;
    strstream server_stream;
    server_stream << server_line.c_str();
    int id, dest_x, dest_y, weight;
    int package_counter = 0;

    // we will check this later
    rPackages.clear();
    while ( server_stream >> id )
    {
        server_stream >> dest_x >> dest_y >> weight;
//        SetPackage( rPackages, id, dest_x, dest_y, currX, currY, weight );
        AddPackage( rPackages, id, dest_x, dest_y, currX, currY, weight );

        package_counter++;
    }
    return package_counter;
}
//---------------------------------------------------------------------------
int Server::FindPackageIndex( Packages& rPackages, int id )
{
    for ( unsigned int i = 0; i < rPackages.size(); i++ )
    {
        if ( rPackages[i].id == id ) return i;
    }
    return -1;
}
//---------------------------------------------------------------------------
void Server::SetPackage( Packages& rPackages, int id, int destX, int destY, int currX, int currY, int weight )
{
    // Check if we know this package
    int pkg_index = FindPackageIndex( rPackages, id );
    if ( pkg_index == - 1 )
    {
        // A new package
        Package tmp;
        tmp.id = id;
        tmp.dest_x = destX;
        tmp.dest_y = destY;
        tmp.last_known_x = currX;
        tmp.last_known_y = currY;
        rPackages.push_back( tmp );
    }
}
//---------------------------------------------------------------------------
void Server::AddPackage( Packages& rPackages, int id, int destX, int destY, int currX, int currY, int weight )
{
    // A new package
    Package tmp;
    tmp.id = id;
    tmp.dest_x = destX;
    tmp.dest_y = destY;
    tmp.last_known_x = currX;
    tmp.last_known_y = currY;
    tmp.weight = weight;
    rPackages.push_back( tmp );
}
//---------------------------------------------------------------------------

