/****************************************************************************
 * The Fifth ICFP Programming Contest entry
 * Copyright (C) 2002  Dragan Tubic
 * tdragan@gel.ulaval.ca
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ****************************************************************************/

#ifndef __CONTEST_H__
#define __CONTEST_H__
//---------------------------------------------------------------------------
#include <stdexcept>
#include <string>
#include <vector>
#include <strstream>
//---------------------------------------------------------------------------
using namespace std;
//---------------------------------------------------------------------------
#define BUFFER_SIZE 16384
//---------------------------------------------------------------------------
class Server;
//---------------------------------------------------------------------------
class Package
{
public:
    int id;
    int dest_x;
    int dest_y;
    int last_known_x;
    int last_known_y;
    int weight;
};
//---------------------------------------------------------------------------
typedef vector<Package> Packages;
//---------------------------------------------------------------------------
class Robot
{
public:
    int pred_x;
    int pred_y;
    int x;
    int y;
    int id;
    int capacity;
    int money;
    Robot();
};
//---------------------------------------------------------------------------
typedef vector<Robot> Robots;
//---------------------------------------------------------------------------
class HomeBase
{
public:
    HomeBase();
    int x;
    int y;
    bool empty;
};
//---------------------------------------------------------------------------
enum SquareType { invalid = -1, empty = 0, wall = 1, lethal = 2 };
//---------------------------------------------------------------------------
class Square
{
public:
    int mGoalDistance;
    int mWallDistance;
    int mType;
    Square();
};
//---------------------------------------------------------------------------
class WorldMap
{
private:
    int mMapSizeX;
    int mMapSizeY;
    Square** mMap;
    std::vector<HomeBase> mHomeBases;
    void AllocateMemory();
    void ReleaseMemory();
public:
    WorldMap();
    WorldMap( unsigned int sizeX, unsigned int sizeY );
    void NewMap( unsigned int sizeX, unsigned int sizeY );
    void NewMap( Server& rServer );
    ~WorldMap();

    Square& operator ()(int x, int y);

    int SizeX() { return mMapSizeX; }
    int SizeY() { return mMapSizeY; }

    HomeBase& GetHomeBase( int index );
    int NoHomeBases();
    int FindClosestHomeBase( int x, int y );
    
    void ComputeWallDistances();
    int FindClosestWall( int x, int y );

    void ComputeGoalDistances( int goalX, int goalY );
    void DoNeighbours( int currDist, int ptX, int ptY, int* new_points_x, int* new_points_y, int& point_counter );

    void MakeStep( int currX, int currY, int& rNewX, int& rNewY );
    float EvaluateStep( int x, int y, int currGoalDist, int currWallDist );
};

//---------------------------------------------------------------------------
class Server
{
private:
    string mHostName;
    int    mHostPort; 
    void Send( string& rMessage );
    void Receive2( string& rMessage );
    char mBuffer[BUFFER_SIZE];
    int mSocket;
public:
    Server();
    ~Server();
    Server( string hostname, int port );
    void Connect();
    void Connect( string hostname, int port );

    void Close();

    void Init( WorldMap& rMap, Robots& rRobots, int& me, Packages& rPackages );

    void ExecuteCommand( string& rCommand );

    int FindRobot( Robots& rRobots, int id );
    void CheckRobot( Robots& rRobots, int id );
    void PositionRobot( Robots& rRobots, int id, int x, int y );
    void MoveRobot( Robots& rRobots, int id, int dx, int dy );
    bool UpdateRobots( Robots& rRobots, Packages& rPackages, Packages& rPickedPackages );
    void ReadLine( string& rMessage );
    void RemoveRobots( Robots& rRobots, vector<int>& rLivingBots );

    int UpdatePackages( Packages& rPackages, int currX, int currY );
    void SetPackage( Packages& rPackages, int id, int destX, int destY, int currX, int currY, int weight );
    void AddPackage( Packages& rPackages, int id, int destX, int destY, int currX, int currY, int weight );
    int FindPackageIndex( Packages& rPackages, int id );

    void LoadPackage( Robots& rRobots, Packages& rPackages, Packages& rPickedPackages, int packId );
    void DropPackage( Robots& rRobots, Packages& rPackages, Packages& rPickedPackages, int packId );
};
//---------------------------------------------------------------------------
#endif //__CONTEST_H__
